// Scanner.java
//
// Informatics 102 Spring 2012
// Code Example
//
// This class implements a scanner for our hypothetical language.  Its job
// is to allow us to iterate through the tokens in a stream of input.  Unlike
// the built-in Scanner class in Java, our scanner will always read one token
// ahead and store it in a couple of fields (one that specifies the kind of
// token it is and another that specifies the actual String that makes up that
// token), while providing a way to ask the Scanner to return the current
// token or to move forward to the next one.

import java.io.IOException;
import java.io.Reader;


public class Scanner
{
	// This is where the Scanner's input comes from.
	private Reader reader;

	// These fields specify the current token (its type and the text itself).
	private TokenType currentTokenType;
	private String currentToken;

	// The nextChar field will always store the next character that should
	// be consumed.  Just as the parser needs one token of look ahead, the
	// scanner needs one character of look ahead in order to make its
	// decisions.
	//
	// If you're wondering why this field's type is int rather than char,
	// it's because we need a special value to indicate that there are no
	// more characters left in the input.  We'll denote that value by the
	// constant NO_MORE_CHARS (-1).
	private int nextChar;
	private static final int NO_MORE_CHARS = -1;
	

	// When we construct a Scanner, we associate it with an input source.
	// Reader is a superclass of many kinds of "readers" (objects that know
	// how to read characters from various input sources) in the Java
	// library.  By passing a Reader rather than something more specific,
	// like a FileReader, we allow our Scanner to work with input coming
	// from any kind of source, rather than a particular one.
	public Scanner(Reader reader)
	{
		this.reader = reader;

		advanceToNextChar();

		// Since our Scanner should always read one token ahead, we'll advance
		// to the next token before returning from the constructor.
		advance();
	}


	// getCurrentTokenType() returns the token type for the current token.
	public TokenType getCurrentTokenType()
	{
		return currentTokenType;
	}
	

	// getCurrentToken() returns the current token.
	public String getCurrentToken()
	{
		return currentToken;
	}
	

	// advance() moves the Scanner forward, so that the current token becomes
	// the next token in the input.  The main job of this method is to read
	// the input one character at a time, eliminating unnecessary whitespace
	// until the beginning of a token is found, then determining where that
	// token ends.
	public void advance()
	{
		// First, we'll consume any whitespace before the beginning of the
		// next token.
		consumeWhitespaceIfAny();

		// Once we've consumed our whitespace, if there are no more characters
		// in the input, we're done; the current token is the special
		// END_OF_INPUT token.
		if (noMoreChars())
		{
			currentTokenType = TokenType.END_OF_INPUT;
			currentToken = "";
			return;
		}

		// If there are characters left in the input, now we need to read the
		// next token and decide what it is.		

		if (nextChar == '{')
		{
			currentTokenType = TokenType.LEFT_BRACE;
			currentToken = "{";
			advanceToNextChar();
			return;
		}
		else if (nextChar == '}')
		{
			currentTokenType = TokenType.RIGHT_BRACE;
			currentToken = "}";
			advanceToNextChar();
			return;
		}
		else if (nextChar == ';')
		{
			currentTokenType = TokenType.SEMICOLON;
			currentToken = ";";
			advanceToNextChar();
			return;
		}
		else if (Character.isLetter(nextChar))
		{
			// If we're sitting on a letter, we should read the entire word
			// where we are -- all of the letters until we find no more
			// letters -- then decide whether it's a reserved word or an
			// identifier.
			
			currentToken = "";
			
			do
			{
				currentToken += (char) nextChar;
				advanceToNextChar();
			}
			while (Character.isLetter(nextChar));
			
			if (currentToken.equals("program"))
			{
				currentTokenType = TokenType.PROGRAM;
				return;
			}
			else if (currentToken.equals("var"))
			{
				currentTokenType = TokenType.VAR;
				return;
			}
			else
			{
				currentTokenType = TokenType.IDENTIFIER;
				return;
			}
		}
		else
		{
			throw new ScanException("Invalid character: " + (char) nextChar);
		}
	}
	
	
	private void consumeWhitespaceIfAny()
	{
		while (!noMoreChars() && Character.isWhitespace(nextChar))
		{
			advanceToNextChar();
		}
	}
	
	
	private void advanceToNextChar()
	{
		try
		{
			nextChar = reader.read();
		}
		catch (IOException e)
		{
			throw new ScanException("IOException occurred", e);
		}
	}


	private boolean noMoreChars()
	{
		return nextChar == NO_MORE_CHARS;
	}
}
